#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "kmeans.h"
#include "linear.h"

/*******************************************************/
/*********** Functions for PCA initialization **********/
/*******************************************************/
MATRIXD computeEigenValues(MATRIXD M, MATRIXD *V){
	long i, j;
	int nrot;
	MATRIXD D;
	
	if (M.row != M.column){
		printf("\nError: Matrix should be square for eigenvalue computation\n");
		exit(1);
	}
	for (i = 0; i < M.row; i++)
		for (j = i+1; j < M.column; j++)
			if (M.data[i][j] != M.data[j][i]){
				printf("\nError: Matrix should be symmetric for eigenvalue computation\n");
				exit(1);
			}
	D = allocateMatrixD(1,M.row);
	*V = allocateMatrixD(M.row,M.row);
	jacobi(M.data,M.row,D.data[0],V->data,&nrot);
	eigsrt(D.data[0],V->data,D.column);
	return D;
}
MATRIXD inverseMatrixD(MATRIXD M){
	MATRIXD inv, temp;
	int i;
	
	if (M.row != M.column){
		printf("\nError: Matrix should be square for inverse computation\n");
		exit(1);
	}
	temp = allocateMatrixD(M.row,M.column);
	copyMatrixD(&temp,M);
	inv = allocateMatrixD(M.row,M.column);
	initializeMatrixD(&inv,0.0);
	for (i = 0; i < M.row; i++)
		inv.data[i][i] = 1;
	gaussj(M.data,M.row,inv.data,inv.column);
	copyMatrixD(&M,temp);
	freeMatrixD(temp);
	return inv;
}
MATRIXD initializeClusterVectorsByPCA(MATRIXD data, int K){
	MATRIXD M, corrData, D, V, newData, newM, temp;
	double d, maxEntry, minEntry, *c;
	long i, j, l, *no;
	
	corrData = computeCorrelationMatrixD(data);
	D = computeEigenValues(corrData,&V);
	newData = multiplyMatrixD(data,V);
	maxEntry = maxMatrixDColumnEntry(newData,0);
	minEntry = minMatrixDColumnEntry(newData,0);
	d = (maxEntry - minEntry) / K;
	c = (double *)malloc((K+1) * sizeof(double));
	c[0] = minEntry;
	for (i = 1; i <= K; i++)
		c[i] = c[i-1] + d;
	
	no = (long *)calloc(K,sizeof(long));
	newM = allocateMatrixD(K,data.column);
	initializeMatrixD(&newM,0.0);
	for (i = 0; i < newData.row; i++)
		for (j = 0; j < K; j++)
			if ((newData.data[i][0] >= c[j]) && (newData.data[i][0] < c[j+1])){
				for (l = 0; l < newM.column; l++)
					newM.data[j][l] += newData.data[i][l];
				no[j]++;
				break;
			}
    
	for (i = 0; i < K; i++)
		for (j = 0; j < newM.column; j++)
			newM.data[i][j] /= no[i];
	temp = inverseMatrixD(V);
	M = multiplyMatrixD(newM,temp);
	
	free(c);
	free(no);
	freeMatrixD(newData);
	freeMatrixD(newM);
	freeMatrixD(corrData);
	freeMatrixD(D);
	freeMatrixD(V);
	freeMatrixD(temp);
	return M;
}
/*******************************************************/
/*********** Functions for kmeans clustering  **********/
/*******************************************************/
MATRIXD kmeans(MATRIXD data, int K, int maxIterNo){
	MATRIX B;
	MATRIXD M, prevM;
	double maxDeltaM;
	int iterNo;
    
    /* it uses PCA for finding the initial vectors */
    M = initializeClusterVectorsByPCA(data,K);
    prevM = allocateMatrixD(K,data.column);
    B = allocateMatrix(data.row,K);
    
    if (maxIterNo < 0)
        maxIterNo = KMEANS_MAX_ITER_NO;
    
    iterNo = 0;
    while (iterNo <= maxIterNo){
	iterNo++;
        copyMatrixD(&prevM,M);
        estimateClustersUsingEuclideanDistance(&B,data,M);
        computeClusterVectors(&M,data,B);
        maxDeltaM = maxAbsoluteDifferenceBetweenMatricesD(M,prevM);
        if (maxDeltaM <= SMALL)
            break;
    }
    
    freeMatrix(B);
    freeMatrixD(prevM);
    return M;
}
void estimateClustersUsingEuclideanDistance(MATRIX *clusterAssigned, MATRIXD data, MATRIXD V){
	long i, j, t, minIndex, K = V.row;
	double *distance = (double *)malloc(V.row * sizeof(double));
    
	initializeMatrix(clusterAssigned,0);
	for (i = 0; i < data.row; i++){
		for (j = 0; j < K; j++){
			distance[j] = 0.0;
			for (t = 0; t < data.column; t++)
				distance[j] += SQRDIST(V.data[j][t],data.data[i][t]);
		}
		minIndex = indexOfMinArrayEntry(distance, K);
		clusterAssigned->data[i][minIndex] = 1;
	}
	free(distance);
}
void computeClusterVectors(MATRIXD *M, MATRIXD data, MATRIX B){
	long i, j, t;
	MATRIX N = allocateMatrix(M->row,1);
    
	initializeMatrixD(M,0.0);
	initializeMatrix(&N,0);
    
	for (t = 0; t < B.row; t++)
		for (i = 0; i < B.column; i++)
			if (B.data[t][i] == 1){
				for (j = 0; j < M->column; j++)
					M->data[i][j] += data.data[t][j];
				N.data[i][0]++;
			}
	for (i = 0; i < N.row; i++)
		if (N.data[i][0] > 0)
			for (j = 0; j < M->column; j++)
				M->data[i][j] /= N.data[i][0];
		else{
			srand((unsigned)time(NULL));
			t = rand() % data.row;
			for (j = 0; j < M->column; j++)
				M->data[i][j] = data.data[t][j];
		}
}
double maxAbsoluteDifferenceBetweenMatricesD(MATRIXD A, MATRIXD B){
	long i, j;
	double maxAbsEntry = fabs(A.data[0][0] - B.data[0][0]);
    
	for (i = 0; i < A.row; i++)
		for (j = 0; j < A.column; j++)
			if (maxAbsEntry < fabs(A.data[i][j] - B.data[i][j]))
				maxAbsEntry = fabs(A.data[i][j] - B.data[i][j]);
	return maxAbsEntry;
}
MATRIX findClusterLabelsUsingEuclideanDistance(MATRIXD data, MATRIXD V){
	long i, j, t, minIndex, K = V.row;
	double *distance = (double *)malloc(V.row * sizeof(double));
	MATRIX labels = allocateMatrix(data.row,1);
    
	for (i = 0; i < data.row; i++){
		for (j = 0; j < K; j++){
			distance[j] = 0.0;
			for (t = 0; t < data.column; t++)
				distance[j] += SQRDIST(V.data[j][t],data.data[i][t]);
		}
		minIndex = indexOfMinArrayEntry(distance, K);
		labels.data[i][0] = minIndex;
	}
	free(distance);
	return labels;
}
int indexOfMinArrayEntry(double A[], int size){
	int minIndex = 0, i;
	for (i = 1; i < size; i++)
		if (A[i] < A[minIndex])
			minIndex = i;
	return minIndex;
}

